#pragma once

#include <string>
#include <map>
#include "ast/type_info.hpp"
#include "symbol_table/symbol_table.hpp"

#define QUANTUM_NULL_VAL QuantumValue()

/* Object for passing context via different levels of (quantum) codegen */
namespace quant_c_ast {
    enum class QuantumOpCode {
        I,
        X,
        Y,
        Z,
        CNOT,
        CCNOT,
        CCNOT_OR,
        H,
        PHASE,
        S,
        T,
        RX,
        RY,
        RZ,
        SWAP,
        ISWAP,
        PSWAP
    };

    class QuantumValue: public symbol_table::Symb {
        public:
            QuantumValue();
            // Quantum memory
            QuantumValue(std::string name,
                         std::pair<int, int> qubitAddress,
                         bool isArg = false
                         );

            // Create "temporary" for sequence of quantum variables
            QuantumValue(QuantumValue& lhs, QuantumValue& rhs);

            // Classical Memory
            QuantumValue(std::string name,
                         type_info::Type classicalType,
                         bool isArg = false,
                         std::string sharedName = ""
                         );

            // Literals
            QuantumValue(std::intmax_t intval, bool isSigned);

            // Members
            std::pair<int, int> qubitAddress;
            type_info::Type classicalType;
            bool isArg;
            bool isClassical;
            bool isShared;
            std::string sharedName;

            // Helper for generating qubit sequences
            std::vector<std::string> qubitAddressSequence;

            // Literal Members
            bool isIntLiteral;
            intmax_t intVal;
            bool isSigned;

            // isNull
            bool isNull;
    };

    enum class OpMod {
        CONTROL,
        INVERSE
    };

    class QuantumOpMod {
        public:
            QuantumOpMod(OpMod modifier, QuantumValue value = QuantumValue());

        // Members
        OpMod modifier;
        QuantumValue value;
    };

    class QuantumContext {
        public:
            QuantumContext(int startQubit);
            void assignQubitAddress(std::string name,
                                    type_info::Type type,
                                    bool isArg = false);
            void declareClassicalMemory(std::string name,
                                        type_info::Type type,
                                        bool isArg = false,
                                        std::string sharedName = "");
            QuantumValue getQuantumValue(std::string name);
            std::string genQUIL();
            // TODO: handle gen for parametric gates
            // QuantumValue buildUnaryClassicalOp();
            // QuantumValue buildBinaryClassicalOp();
            void buildMove(QuantumValue memoryVal, QuantumValue storeVal);
            void buildQuantumInit(QuantumValue memoryVal, QuantumValue storeVal);
            QuantumValue buildCircuitCall(std::string name, std::vector<QuantumValue> args);
            QuantumValue buildMeasurement(QuantumValue qubits);
            QuantumValue buildUnaryGateOp(QuantumOpCode op,
                                  QuantumValue operand);
            QuantumValue buildBinaryGateOp(QuantumOpCode op,
                                   QuantumValue left,
                                   QuantumValue right);
            void buildReturn(QuantumValue returnVal);
            void setFunctionName(std::string name);
            void addFunctionArgument(std::string name, type_info::Type type);
            void setFunctionReturnType(type_info::Type type);
            void insertOpMod(QuantumOpMod mod);
            QuantumValue generateTempMemory(type_info::Type type);

            // Members
            std::string functionName;
            std::map<std::string, QuantumValue> functionArgs;
            type_info::Type returnType;
            int tmpId = 0;
            std::map<std::string, QuantumValue> declaredClassicalMemory;
            std::vector<QuantumOpMod> nextMods;
            std::vector<std::string> instructions;
            int startingQubit;
            int nextUnassignedQubit;
            symbol_table::SymbolTable<QuantumValue> qubitAddresses;
    };
}
