#include "ast/type_info.hpp"
#include <vector>
#include <string>
#include <climits>
#include "llvm/IR/Argument.h"

/**
 * A class for storing type information about symbols.
 */

using namespace type_info;

// Default constructor
Type::Type() {
    value = TypeValue::UNKNOWN; isSigned = true;
    quality = TypeQualityVal::UNKNOWN; storage = StorageSpecVal::UNKNOWN;
    isScalar = true;
    subType = nullptr;
    isArray = false; size = 1;
    isFunction = false; functionArgTypes = std::vector<Type>();
    isVarArg = false;
    isPointer = false;
}

Type::Type(const Type& other) {
    value = other.value; isSigned = other.isSigned;
    quality = other.quality; storage = other.storage;
    isScalar = other.isScalar;
    subType = (other.subType == nullptr) ?
               nullptr :
               std::make_shared<Type>(*(other.subType));
    isArray = other.isArray; size = other.size;
    isFunction = other.isFunction; functionArgTypes = other.functionArgTypes;
    isVarArg = other.isVarArg;
    isPointer = other.isPointer;
}

// Constructor for scalar types
Type::Type(TypeValue typeVal, bool signedVal,
           TypeQualityVal qualityVal, StorageSpecVal storageVal) {
    value = typeVal; isSigned = signedVal;
    quality = qualityVal; storage = storageVal;
    isScalar = true;
    subType = nullptr;
    isArray = false; size = 1;
    isFunction = false; functionArgTypes = std::vector<Type>();
    isVarArg = false;
    isPointer = false;
}

// Constructor for arrays
Type::Type(std::shared_ptr<Type> typeVal, int arraySize,
           TypeQualityVal qualityVal, StorageSpecVal storageVal) {
    value = TypeValue::VOID; isSigned = false;
    quality = qualityVal; storage = storageVal;
    isScalar = false;
    subType = std::make_shared<Type>(*typeVal);
    isArray = true; size = arraySize;
    isFunction = false; functionArgTypes = std::vector<Type>();
    isVarArg = false;
    isPointer = false;
}

// Constructor for functions
Type::Type(std::shared_ptr<Type> typeVal,
           std::vector<Type> argTypes,
           bool varArg,
           TypeQualityVal qualityVal, StorageSpecVal storageVal) {
    value = TypeValue::VOID; isSigned = false;
    quality = qualityVal; storage = storageVal;
    isScalar = false;
    subType = std::make_shared<Type>(*typeVal);
    isArray = false; size=1;
    isFunction = true; functionArgTypes = argTypes;
    isVarArg = varArg;
    isPointer = false;
}

// Constructor for pointers
Type::Type(std::shared_ptr<Type> typeVal,
           TypeQualityVal qualityVal,
           StorageSpecVal storageVal) {
    value = TypeValue::VOID; isSigned = false;
    quality = qualityVal; storage = storageVal;
    isScalar = false;
    subType = std::make_shared<Type>(*typeVal);
    isArray = false; size=1;
    isFunction = false; functionArgTypes = std::vector<Type>();
    subType = typeVal;
    isVarArg = false;
    isPointer = true;
}

// Return string for value
std::string Type::get_value_string() const {
    switch(value) {
        case TypeValue::CHARACTER : return "character";
        case TypeValue::SHORT     : return "short";
        case TypeValue::INTEGER   : return "integer";
        case TypeValue::LONG      : return "long";
        case TypeValue::LONGLONG  : return "long long";
        case TypeValue::FLOAT     : return "float";
        case TypeValue::DOUBLE    : return "double";
        case TypeValue::BOOLEAN   : return "boolean";
        case TypeValue::ERROR     : return "error";
        case TypeValue::UNKNOWN   : return "unknown";
        case TypeValue::VOID      : return "void";
        default: return "notype";
    }
}

// Return string for quality
std::string Type::get_quality_string() const {
    switch(quality) {
        case TypeQualityVal::VOLATILE       : return "volatile";
        case TypeQualityVal::CONST          : return "const";
        case TypeQualityVal::VOLATILE_CONST : return "volatile const";
        case TypeQualityVal::NONE           : return "";
        case TypeQualityVal::UNKNOWN        : return "unknown";
        default: return "";
    }
}

// Return string for quality
std::string Type::get_storage_string() const {
    switch(storage) {
        case StorageSpecVal::TYPEDEF  : return "typedef";
        case StorageSpecVal::EXTERN   : return "extern";
        case StorageSpecVal::STATIC   : return "static";
        case StorageSpecVal::REGISTER : return "register";
        case StorageSpecVal::QUANTUM  : return "quantum";
        case StorageSpecVal::AUTO     : return "";
        case StorageSpecVal::UNKNOWN  : return "unknown";
        default: return "";
    }
}

// Get base values for complex types
std::pair<Type *, bool>Type::get_base_type() {
    if (!isScalar && (isFunction || isArray || isPointer)) {
        if (subType == nullptr || subType->value == TypeValue::UNKNOWN)
            return std::make_pair(this, false);
        else return subType->get_base_type();
    }
    else {
        return std::make_pair(this, true);
    }
}
TypeValue Type::get_base_type_value() const {
    if (subType != nullptr) {
        return subType->get_base_type_value();
    } else {
        return value;
    }
}
TypeQualityVal Type::get_base_type_quality() const {
    if (subType != nullptr) {
        return subType->get_base_type_quality();
    } else {
        return quality;
    }
}
StorageSpecVal Type::get_base_type_storage() const {
    if (subType != nullptr) {
        return subType->get_base_type_storage();
    } else {
        return storage;
    }
}

void Type::set_base_type(Type newBase) {
    if (!isScalar && (isFunction || isArray || isPointer) &&
         subType.get() == nullptr) {
        subType = std::make_shared<Type>();
        *subType = newBase;
    }
    else if (isScalar) {
        *(this) = newBase;
    }
    else {
        subType->set_base_type(newBase);
    }
}

llvm::Type * Type::codegen_type(llvm::LLVMContext& context) {
    const int byte_width = CHAR_BIT;
    if (isScalar) {
        switch(value) {
            case TypeValue::CHARACTER:
                return llvm::IntegerType::get(context, sizeof(char)*byte_width);
            case TypeValue::SHORT:
                return llvm::IntegerType::get(context, sizeof(short)*byte_width);
            case TypeValue::INTEGER:
                return llvm::IntegerType::get(context, sizeof(int)*byte_width);
            case TypeValue::LONG:
                return llvm::IntegerType::get(context, sizeof(long)*byte_width);
            case TypeValue::LONGLONG:
                return llvm::IntegerType::get(context,
                                              sizeof(long long)*byte_width);
            case TypeValue::FLOAT:     return llvm::Type::getFloatTy(context);
            case TypeValue::DOUBLE:    return llvm::Type::getDoubleTy(context);
            case TypeValue::BOOLEAN:
                return llvm::IntegerType::get(context, sizeof(bool)*byte_width);
            case TypeValue::VOID:     return llvm::Type::getVoidTy(context);
            case TypeValue::CUSTOM:
            case TypeValue::ERROR:
            case TypeValue::UNKNOWN:
            default : return nullptr;
        }
    }
    else if (isFunction) {
        std::vector<llvm::Type *> arg_type_values;
        for (auto& argType : functionArgTypes) {
            llvm::Type* argTypeVal = argType.codegen_type(context);
            if (argTypeVal->isArrayTy()) {
                // Convert array to pointer for arguments
                argTypeVal = llvm::PointerType::getUnqual(argTypeVal->getContainedType(0));
            }
            arg_type_values.push_back(argTypeVal);
        }
        llvm::ArrayRef<llvm::Type *> argTypeArray = arg_type_values;
        llvm::Type* returnType = subType->codegen_type(context);
        return llvm::FunctionType::get(returnType,
                                       argTypeArray,
                                       isVarArg);
    }
    else if (isArray) {
        return llvm::ArrayType::get(subType->codegen_type(context), size);
    }
    else if (isPointer) {
        auto pointeeType = subType->codegen_type(context);
        return llvm::PointerType::getUnqual(pointeeType);
    }
    else {
        return nullptr; // Error
    }
}

std::pair<std::string, int> Type::quantum_codegen_type() {
    const int byte_width = CHAR_BIT;
    const bool isOctetWidth = byte_width == 8;
    if (isScalar) {
        switch(value) {
            case TypeValue::CHARACTER: {
                if(isOctetWidth) {
                    return std::make_pair("OCTET", 1);
                }
                else {
                    return std::make_pair("BIT", byte_width);
                }
            }
            case TypeValue::SHORT: {
                if(sizeof(short) == sizeof(int)) {
                    return std::make_pair("INTEGER", 1);
                }
                else if (isOctetWidth) {
                    return std::make_pair("OCTET", sizeof(short));
                }
                else {
                    return std::make_pair("BIT", byte_width*sizeof(short));
                }
            }
            case TypeValue::INTEGER:
                return std::make_pair("INTEGER", 1);
            case TypeValue::LONG: {
                if(sizeof(long) == sizeof(int)) {
                    return std::make_pair("INTEGER", 1);
                }
                else if (isOctetWidth) {
                    return std::make_pair("OCTET", sizeof(long));
                }
                else {
                    return std::make_pair("BIT", byte_width*sizeof(long));
                }
            }
            case TypeValue::LONGLONG: {
                if(sizeof(long long) == sizeof(int)) {
                    return std::make_pair("INTEGER", 1);
                }
                else if (isOctetWidth) {
                    return std::make_pair("OCTET", sizeof(long long));
                }
                else {
                    return std::make_pair("BIT", byte_width*sizeof(long long));
                }
            }
            case TypeValue::FLOAT:     return std::make_pair("REAL", 1);
            case TypeValue::DOUBLE:    return std::make_pair("REAL", 1);
            case TypeValue::BOOLEAN:   return std::make_pair("BIT", 1);
            default : return std::make_pair("ERROR", 1);
        }
    }
    else if (isArray) {
        auto subTypeGen = subType->quantum_codegen_type();
        return std::make_pair(subTypeGen.first, subTypeGen.second*size);
    }
    else {
        return std::make_pair("ERROR", 1); // Error
    }
}

int Type::getTypeBitSize(Type typeVal) {
    auto [type, size] = typeVal.quantum_codegen_type();

    if (type == std::string("OCTET")) {
        return CHAR_BIT*size;
    }
    else if (type == std::string("INTEGER")) {
        return CHAR_BIT*sizeof(int)*size;
    }
    else if (type == std::string("REAL")) {
        return CHAR_BIT*sizeof(float)*size;
    }
    else {
        return size;
    }
}

bool Type::typeValCompare(Type other) {
    if (*this == other) return true;
    else {
        if (isScalar) {
            return this->value == other.value;
        }
        else {
            bool structureMatch= this->isArray == other.isArray &&
                   this->isFunction == other.isFunction &&
                   this->isPointer == other.isPointer &&
                   this->subType;
            bool subTypesMatch = true;
            if ((this->subType.get() == nullptr &&
                 other.subType.get() != nullptr) ||
               (this->subType.get() != nullptr &&
                other.subType.get() == nullptr)) {
                    return false;
            }
            else if (this->subType.get() != nullptr &&
                     other.subType.get() != nullptr) {
                subTypesMatch = this->subType->typeValCompare(*(other.subType));
            }
            return subTypesMatch && structureMatch;
        }
    }
}

// Return string for complete type info (includes array and function info)
Type::operator std::string() const {
    std::string typeString;

    // Pointer:
    // Add pointer description and
    // use string of subType as pointee type
    if (isPointer) {
        if (subType == nullptr) {
            typeString += " error ";
        }
        else {
            typeString += " ( " + std::string(*subType) + " ) ";
        }

        typeString += "* ";
        typeString += this->get_quality_string();
    }

    // Function:
    // Build string for argument types
    // and use subType string as return type
    else if (isFunction) {
        if (subType == nullptr) {
            typeString += " error ";
        }
        else {
        typeString += std::string(*subType);
        }

        typeString += "( ";
        for (auto& argType : functionArgTypes) {
            typeString += std::string(argType);
            typeString += ", ";
        }
        // Erase last comma
        typeString.pop_back();
        typeString.pop_back();

        if(isVarArg) {
            typeString += ", ...";
        }

        typeString += " ) ";


    }

    // Array
    // Use string of subType as element type and
    // append size information
    else if (isArray) {
        if (subType == nullptr) {
            typeString += " error ";
        }
        else {
        typeString += std::string(*subType);
        }

        if (size > 0) {
            typeString += " [ ";
            typeString +=  std::to_string(size);
            typeString += " ]";
        }
        else {
            typeString += " [ ]";
        }
    }

    // Scalar type:
    // Use basic type info to build the type string
    else {
        typeString += this->get_storage_string();
        typeString += " ";
        typeString += this->get_quality_string();
        typeString += " ";
        typeString += isSigned ? "signed" : "unsigned";
        typeString += " ";
        typeString += this->get_value_string();
    }

    return typeString;
}

bool type_info::operator ==(const Type& lhs, const Type& rhs) {
    bool bases_match = lhs.value == rhs.value &&
           lhs.isSigned == rhs.isSigned &&
           lhs.quality == rhs.quality &&
           lhs.storage == rhs.storage &&
           lhs.size == rhs.size &&
           lhs.isArray == rhs.isArray &&
           lhs.isFunction == rhs.isFunction &&
           lhs.isPointer == rhs.isPointer &&
           lhs.functionArgTypes == rhs.functionArgTypes
           ;
    bool subTypes_match = true;
    if ((lhs.subType.get() == nullptr && rhs.subType.get() != nullptr) ||
        (lhs.subType.get() != nullptr && rhs.subType.get() == nullptr)) {
            return false;
    }
    else if (lhs.subType.get() != nullptr && rhs.subType.get() != nullptr) {
        subTypes_match = (*lhs.subType) == (*rhs.subType);
    }

    return bases_match && subTypes_match;
}

bool type_info::operator !=(const Type& lhs, const Type& rhs) {
    return !(lhs == rhs);
}
