#include <iostream>
#include <iomanip>
#include <algorithm>
#include <iterator>
#include <string_view>
#include <cmath>
#include <climits>
#include <exception>
#include <stdexcept>
#include "symbol_table/symbol_table.hpp"
#include "ast/ast.hpp"
#include "quant_c_parse.hpp"

using namespace quant_c_ast;
using namespace symbol_table;
using namespace type_info;
using namespace quant_c_compiler;

// Static initialization of ASTNode
SymbolTable<SymbolInfo> ASTNode::semanticTable = SymbolTable<SymbolInfo>();
SymbolTable<SymbolValue> ASTNode::codegenTable = SymbolTable<SymbolValue>();
int ASTNode::semanticsErrors = 0;
int ASTNode::qubitOffset = 0;
int ASTNode::numQuantumFuncs = 0;
bool ASTNode::firstQuantFunc = true;
std::string ASTNode::quilCode = "";
llvm::LLVMContext ASTNode::LLVMContext;
llvm::IRBuilder<> ASTNode::Builder(LLVMContext);
std::unique_ptr<llvm::Module> ASTNode::LLVMModule;
// llvm::ValueSymbolTable ASTNode::NamedValues;

void ASTNode::genIndent(int indent) {
    for (int i = 0; i <= indent; i += 1) {
        std::cout << "    ";
    }
} // genIndent

void ASTNode::mustBe(bool assertion, std::string errorMsg) {
    if (! assertion) {
        std::cout << errorMsg << "\n";
        semanticsErrors++;
    }
} // mustBe

void ASTNode::mustNotBe(bool assertion, std::string errorMsg) {
    if (assertion) {
        std::cout << errorMsg << "\n";
        semanticsErrors++;
    }
} // mustNotBe

void ASTNode::storageMustBe(type_info::Type testType, type_info::StorageSpecVal requiredStorage, std::string errorMsg) {
    if (testType.storage != requiredStorage) {
        std::cout << errorMsg << "\n";
        semanticsErrors++;
    }
}
void ASTNode::storageMustBe(type_info::Type testType, std::vector<StorageSpecVal> requiredStorage, std::string errorMsg) {
    bool not_found = (std::find_if(requiredStorage.begin(),
                                   requiredStorage.end(),
                                   [&](StorageSpecVal s){
                                       return s == testType.storage;
                                   })
                                   ==     requiredStorage.end());
    if (not_found) {
        std::cout << errorMsg << "\n";
        semanticsErrors++;
    }
}

void ASTNode::qualityMustBe(type_info::Type testType, TypeQualityVal requiredQuality, std::string errorMsg) {
    if (testType.quality != requiredQuality) {
        std::cout << errorMsg << "\n";
        semanticsErrors++;
    }
}
void ASTNode::qualityMustBe(type_info::Type testType, std::vector<TypeQualityVal> requiredQuality, std::string errorMsg) {
    bool not_found = (std::find_if(requiredQuality.begin(),
                                requiredQuality.end(),
                                [&](TypeQualityVal q) {
                                    return q == testType.quality;
                                })
                                == requiredQuality.end());
    if (not_found) {
        std::cout << errorMsg << "\n";
        semanticsErrors++;
    }
}


void ASTNode::typeMustBe(Type testType, TypeValue requiredType, std::string errorMsg) {
    if ((testType.value == TypeValue::ERROR) ||
        (testType.value != requiredType)) {
        std::cout << errorMsg << "\n";
        semanticsErrors++;
    }
} // typeMustBe

void ASTNode::typeMustBe(Type testType, std::vector<TypeValue> requiredType, std::string errorMsg) {
    bool not_found = (std::find_if(requiredType.begin(),
                                   requiredType.end(),
                                   [&](Type t){
                                       return t == testType.value;
                                   })
                                   ==     requiredType.end());
    if (not_found) {
        std::cout << errorMsg << "\n";
        semanticsErrors++;
    }
} // typeMustBe

void ASTNode::typeMustNotBe(Type testType, TypeValue invalidType, std::string errorMsg) {
    if ((testType.value == TypeValue::ERROR) ||
        (testType.value == invalidType)) {
        std::cout << errorMsg << "\n";
        semanticsErrors++;
    }
} // typeMustNotBe

void ASTNode::typeMustNotBe(Type testType, std::vector<TypeValue> invalidType, std::string errorMsg) {
    bool invalid_type_found = (std::find_if(invalidType.begin(),
                                        invalidType.end(),
                                        [&](Type t){
                                            return t == testType.value;
                                        })
                                        != invalidType.end());
    if ((testType.value == TypeValue::ERROR) || invalid_type_found) {
        std::cout << errorMsg << "\n";
        semanticsErrors++;
    }
} // typeMustNotBe

void ASTNode::typesMustBeEqual(Type type1, Type type2, std::string errorMsg) {
    if ((type1.value == TypeValue::ERROR) ||
        (type2.value == TypeValue::ERROR) ||
        (type1.value != type2.value)) {
        std::cout << errorMsg << "\n";
        semanticsErrors++;
    }
} // typesMustBeEqual

void ASTNode::typesMustBeSimilar(Type type1, Type type2, std::vector<TypeValue> validType, std::string errorMsg) {
    bool first_type_invalid = (std::find_if(validType.begin(),
                                          validType.end(),
                                          [&](TypeValue t){
                                              return t == type1.value;
                                          })
                                          == validType.end());
    bool second_type_invalid = (std::find_if(validType.begin(),
                                          validType.end(),
                                          [&](TypeValue t){
                                              return t == type2.value;
                                          })
                                          == validType.end());
    if ( first_type_invalid || second_type_invalid ) {
        std::cout << errorMsg << "\n";
        semanticsErrors++;
    }
} // typesMustBeEqual

std::string ASTNode::error() {
    return std::string("Error (line ") + std::to_string(linenum) + "): ";
} // error

ASTNode::ASTNode(ASTNodeKind nodeKind) {
    this->nodeKind = nodeKind;
    linenum = -1;
    colnum = -1;
} // ASTNode()

ASTNode::ASTNode(ASTNodeKind nodeKind, int line, int col) {
    this->nodeKind = nodeKind;
    linenum = line;
    colnum = col;
} // ASTNode(line, col)

bool ASTNode::isNull() {
    return false; // often redefined in a subclass
} // isNull()

void ASTNode::unparse(int indent) {
    // This routine is normally redefined in a subclass
} // unparse()

void ASTNode::checkSemantics() {
    // This routine is normally redefined in a subclass
} // checkSemantics()

llvm::Value* ASTNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    return nullptr;
} // codegen()

QuantumValue ASTNode::quantum_codegen(QuantumContext& context) {
    return QUANTUM_NULL_VAL;
}// quantum_codegen()

// class ASTNode


NullNode::NullNode() : ASTNode::ASTNode(ASTNodeKind::AST_NullNode) {}

bool NullNode::isNull() {
    return true;
}

void NullNode::unparse(int indent) {
    // no action
}
// class NullNode

TranslationUnitNode::TranslationUnitNode() :
ASTNode::ASTNode(ASTNodeKind::AST_TranslationUnitNode) {
    this->declList = {};
}

TranslationUnitNode::TranslationUnitNode(std::vector<std::shared_ptr<DeclNode>>, int line, int col) :
ASTNode::ASTNode(ASTNodeKind::AST_TranslationUnitNode, line, col) {
    this->declList = declList;
}

void TranslationUnitNode::setName(std::string name) {
    module_name = name;
}

void TranslationUnitNode::unparse(int indent) {
    for (auto& declNode : declList) {
        declNode->unparse(indent);
    }
}
bool TranslationUnitNode::classof(const ASTNode * node) {
    return node->getNodeKind() == ASTNodeKind::AST_TranslationUnitNode;
} // classof
void TranslationUnitNode::checkSemantics() {
    semanticTable.openScope(); // Open global scope

    // Check declarations
    if (declList.empty()) {
        mustBe(false, error() + "Program has no content.");
    }
    for (auto& declNode : declList) {
        auto declNodepPtr = declNode.get();
        if (llvm::isa<DeclarationNode>(declNodepPtr)) {
            DeclarationNode* decl = llvm::cast<DeclarationNode>(declNodepPtr);
            decl->isGlobalDec = true;
        }
        declNode->checkSemantics();
    }

    try {
        semanticTable.closeScope();
    } catch (EmptySTException &emptyEx){
        // can't happen
    }

} // checkSemantics

bool TranslationUnitNode::isSemanticsCorrect() {
    checkSemantics();
    return (semanticsErrors == 0);
} // isSemanticsCorrect

llvm::Value* TranslationUnitNode::codegen(llvm::Value* parent, llvm::Type* typeInfo) {
    // Initalize module for translation unit
    LLVMModule = std::make_unique<llvm::Module>(module_name, LLVMContext);

    // Insert signature of quantum communication function
    llvm::Type * quant_comms_return_type = llvm::Type::getIntNTy(LLVMContext,
                                                        sizeof(int)*CHAR_BIT);
    std::vector<llvm::Type *> args_types;
    args_types.push_back(llvm::Type::getIntNPtrTy(LLVMContext,
                                                  sizeof(char)*CHAR_BIT));
    args_types.push_back(llvm::Type::getIntNPtrTy(LLVMContext,
                                                  sizeof(char)*CHAR_BIT));
    llvm::ArrayRef<llvm::Type *> quantum_comms_args = args_types;
    bool isVarArg = false;
    llvm::FunctionType* quant_comm_func_type = llvm::FunctionType::get(
                            quant_comms_return_type,
                            quantum_comms_args,
                            isVarArg
                            );
    llvm::Function * quant_comm_func = llvm::Function::Create(
                                            quant_comm_func_type,
                                            llvm::Function::ExternalLinkage,
                                            "quant_com",
                                            LLVMModule.get());
    quant_comm_func->setCallingConv(llvm::CallingConv::C);

    // Set argument names in function code from declarator
    llvm::Argument* func_arg = quant_comm_func->getArg(0);
    func_arg->setName("quil_program");

    // Generate code for declarations
    for (auto& decl : declList) {
        //llvm::Value* decl_code =
        decl->codegen();
    }

    /*
    // TODO: look into optimizations
    auto optimizer = std::make_unique<llvm::legacy::FunctionPassManager>(LLVMModule.get());
    optimizer->add(llvm::createInstructionCombiningPass());
    */

    return nullptr;
}

// class TranslationUnitNode
