#include <iostream>
#include <fstream>
#include <filesystem>
#include "token/token.hpp"
#include "ast/ast.hpp"
#include "scanner_shim.hpp"
#include "quant_c_parse.hpp"

#define MIN_ARGC 2
#define EOF_TOKEN 0

using namespace quant_c_token;
using namespace quant_c_compiler;
using namespace quant_c_ast;

int main(int argc, char *argv[]) {

    if (argc < MIN_ARGC) {
        std::cerr << "Expected one argument; "
                  << "the name of a file to generate code for"
                  << std::endl;
        return -1;
    }

    // Open the file to scan
    std::cout << "Reading file\n";
    std::ifstream quant_c_file;
    try {
        quant_c_file.open(argv[1]);
    } catch (...) {
        std::cerr << "Unable to open file \"" << argv[1] << "\"\n";
        return -1;
    }

    // Create the parser and parse the file
    std::cout << "Creating scanner\n";
    Scanner test_scanner(quant_c_file, std::cout);
    std::cout << "Creating parser\n";
    TranslationUnitNode ast;
    Parser test_parser(test_scanner, ast);
    int parse_result = test_parser.parse();

    // Check prorgam semantics and generate code if succesfully parsed
    if (parse_result == 0) {
        std::cout << "Parsing successfully completed.\n";

        // Verify semantics
        std::cout << "Checking semantics...\n";
        if(!(ast.isSemanticsCorrect())) {
            std::cout << "Program failed to compile due to one or more errors.\n";
            return -1;
        }
        std::cout << "Semantics verified; no issues were detected\n"
                  << "Generating code\n";

        // Generate name of module
        std::string cd = std::filesystem::current_path();
        std::filesystem::path original_path(argv[1]);
        ast.setName(std::string(original_path.stem()));
        ast.codegen();
        // TODO: Check if return value from TranslationUnitNode is meaningful
        // Pull IR from Module into string stream, then write to file
        std::string output_file_name = cd + "/" +
                                       std::string(original_path.stem()) +
                                       ".ll";
        std::string output_content;
        llvm::raw_string_ostream instruction_stream(output_content);
        instruction_stream << *(ASTNode::LLVMModule);
        instruction_stream.flush();
        std::fstream output_file(output_file_name, std::fstream::out);
        output_file << output_content;
        output_file.close();
    }
    else {
        std::cout << "Error code " << parse_result << ": Parsing ended early due to errors.\n";
    }

    // Close file
    try {
        quant_c_file.close();
    } catch (...) {
        std::cerr << "Error while closing file \"" << argv[1] << "\"\n";
        return -1;
    }

    return 0;
}
