int printf ( const char * format, ... );

// Qubits may only be passed by reference.
void quantum toffoli_001(quantum bool(&x)[3], quantum bool &y) {
        // Negate qubits 0 and 1
        // Precedence: [] before %X
        %X x[0]; %X x[1];

        // "+C" is an quantum operation modifier on %X (NOT gate);
        //    it connects the qubits of x as control qubits to NOT
        %X +C(x) y;
        %X x[0]; %X x[1]; // Uncompute; same precedence as first line
} // toffoli_001

// The other Toffoli gate functions follow a similar pattern, except
// for how the qubits of x control the NOT gate
// (i.e. are negated before control)
void quantum toffoli_101(quantum bool(&x)[3], quantum bool &y) {
        %X x[1];
        %X +C(x) y;
        %X x[1]; // Uncompute
} // toffoli_101

void quantum toffoli_010(quantum bool(&x)[3], quantum bool &y) {
        %X x[0]; %X x[2];
        %X +C(x) y;
        %X x[0]; %X x[2]; // Uncompute
} // toffoli_010

void quantum toffoli_100(quantum bool(&x)[3], quantum bool &y) {
        %X x[1]; %X x[2];
        %X +C(x) y;
        %X x[1]; %X x[2]; // Uncompute
} // toffoli_100

// Example Oracle: s = 110 (binary)
void quantum black_box(quantum bool(&x)[3], quantum bool(&y)[3]) {
        toffoli_001(x, y[0]);
        toffoli_101(x, y[0]);
        toffoli_010(x, y[1]);
        toffoli_100(x, y[1]);
        toffoli_001(x, y[2]);

        %X +C(x) y[2];
} // black_box

int quantum simons_algorithm_step() {
        quantum bool x[3] = {0,0,0};
        quantum bool y[3] = {0,0,0};
        int result;

        %H x;
        black_box(x, y);
        %M y;
        %H x;

        // Determine result by measuring x
        result = %M x;
        return result;
} // simons_algorithm_step


// From this point on, program is classical code and should
// thus follow standard C precedence
int classic_black_box(int x) {
        switch(x) {
                case 0: case 6:
                        return 0;
                case 1: case 7:
                        return 1;
                case 2: case 4:
                        return 2;
                case 3: case 5:
                        return 4;
                default: return -1; // error
        }
} // classic_black_box

// Attempt to solve the system of equations as a mtrix using
// Gaussian Elimination and back substitution
void solveMatrix_mod2(int a[2][3]) {
    int m = 2, n = 3;

    // Gaussian Elimination mod 2
    int row, column, rows_after;
    for(row=0; row<m-1; row++){
        for(rows_after=row+1; rows_after<m; rows_after++) {
            int term = a[rows_after][row];
            for(column=0; column<n; column++) {
                int value = (a[rows_after][column] - term * a[row][column]) % 2;
                a[rows_after][column] = (value == -1) ? 1 : value;
            } // column
        } // rows_after
    } // row

    // Back substitution mod 2
    for(row=m-1; row>=0; row--) {
        for(rows_after=row-1; rows_after>=0; rows_after--) {
            int term = a[rows_after][row];
            for(column=n-1; column>0; column--) {
                int value = (a[rows_after][column] - term * a[row][column]) % 2;
                a[rows_after][column] = (value == -1) ? 1 : value;
            } // column
        } // rows_after
    } // rows

} // solveMatrix_mod2()

#define THIRD_BIT_MASK 0x4
#define SECOND_BIT_MASK 0x2
#define FIRST_BIT_MASK 0x1

int main() {
        // Since we are using a 3 bit in/out oracle, perform the quantum step twice
        int s = 0;
        int i;

        int step1_result = simons_algorithm_step();
        int step2_result = simons_algorithm_step();

        // Solve system of equations to determine all but the last bit of s
        int matrix[2][3];
        if (step2_result > step1_result) { // Swap rows
            int temp = step1_result;
            step1_result = step2_result;
            step2_result = temp;
        }
        matrix[0][0] = (step1_result & THIRD_BIT_MASK) >> 2;
        matrix[0][1] = (step1_result & SECOND_BIT_MASK) >> 1;
        matrix[0][2] = step1_result & FIRST_BIT_MASK;
        matrix[1][0] = (step2_result & THIRD_BIT_MASK) >> 2;
        matrix[1][1] = (step2_result & SECOND_BIT_MASK) >> 1;
        matrix[1][2] = step2_result & FIRST_BIT_MASK;
        // Current compiler causes SEGFAULT when this functin is ran
        solveMatrix_mod2(matrix);

        // For this case, easiest guess is top row of solved matrix
        // Take all but the last bit
        for (i = 0; i < 1; i++) {
                s = (s ^ matrix[0][i]) << 1;
        }
        s <<= 1;

        // Check result with classical version of oracle to decide guess for last bit
        if (classic_black_box(s) != classic_black_box(0)) {
                s += 1;
        }

        // Print guess (probability of guessing right > 1/2)
        printf("guess for 's' = %d\n", s);
        return 0;
}
